class ConfigurationFileError(Exception):
    """Base exception class for errors relating to configuration files."""


class BlankKeyError(ConfigurationFileError):
    """Raised when a key is left blank in the config file."""


class SectionNotFoundError(ConfigurationFileError):
    """Raised when a section is missing from the config file."""


class RequiredKeyMissingError(ConfigurationFileError):
    """Raised when a required key is missing from the config fie."""


class NotificationsNotConfiguredError(ConfigurationFileError):
    """Raised when neither notification option is set."""


class SlackCfgError(ConfigurationFileError):
    """Base exception class for errors relation to the Slack Configuration."""


class EmailAuthenticationError(ConfigurationFileError):
    """Raised when either `email_from` or `password` are incorrect in config file."""


class IndeedAuthenticationError(ConfigurationFileError):
    """Raised when there is an issue with the Indeed publisher key."""



from .exceptions import (
    BlankKeyError,
    ConfigurationFileError,
    RequiredKeyMissingError,
    SectionNotFoundError,
    SlackCfgError,
)
from .jobnotify import (
    build_url,
    construct_email,
    construct_slack_message,
    email_notify,
    INDEED_API_LIMIT,
    indeed_api_request,
    INDEED_BASE_URL,
    jobnotify,
    notify,
    send_email,
    slack_notify,
)
from .utils import (
    EmailMatch,
    get_sanitised_params,
    get_section_configs,
    initial_setup,
    load_json_db,
    load_cfg,
    process_args,
)



#!/usr/bin/env python3
from configparser import DuplicateOptionError
import email
import json
import logging
import os
import re
import smtplib
import sys
from urllib.parse import urlencode
import urllib.request

from slackclient import SlackClient

from .exceptions import (
    ConfigurationFileError,
    EmailAuthenticationError,
    IndeedAuthenticationError,
    SlackCfgError,
)
from .utils import (
    get_sanitised_params,
    get_section_configs,
    initial_setup,
    load_json_db,
    process_args,
    write_json_db,
)

INDEED_BASE_URL = 'http://api.indeed.com/ads/apisearch'
INDEED_API_LIMIT = 25
DB_DIR = os.path.join(os.path.expanduser('~'), '.jobnotify', 'databases')
PATH_TO_CFG = os.path.join(os.path.expanduser('~'), '.jobnotify', 'jobnotify.config')


def build_url(base_url, params):
    """Return a correctly formatted URL.

    Args:
        base_url: base URL of the API
        params: dictionary of parameters to be encoded.

    Returns:
        URL string
    """
    encoded_params = urlencode(params)
    return f'{base_url}?{encoded_params}'


def construct_slack_message(posts):
    """Construct a Slack message for sending.

    The Slack RTM API (https://api.slack.com/rtm#limits)
    recommends that a single message be no longer than 4 kB.
    To handle this, we construct our message, and then
    subsequently split into messages of 10 listings or less.

    An iterable is returned: either a single element list
    containing a message with <= 10 listings, or a
    generator with > 10 listings.

    Args:
        posts: dictionary containing new job listings.

    Returns:
        msg_it: an iterable containing the message(s) to
            be posted. Note this may be a single element
            list or a generator.
    """
    nposts = len(posts)

    # build the full message
    msg_template = '{}. <{url}|{jobtitle} @ {company}>\nSnippet: {desc}\n'
    msg = '\n'.join(msg_template.format(i+1, **p) for i, p in enumerate(posts.values()))

    if nposts > 10:
        logging.debug('Splitting message into %d chunks..', (nposts//10)+1)
        # split the message after 10 listings, i.e., on a `11.`, `21.`, etc.
        t = [''] + re.split(r'(\d?\d1\.)', msg)
        # create an iterator from the above list
        it = iter(t)
        # create a generator which pairs successive elements of the original list
        msg_it = (m+next(it, '') for m in it)
    else:
        msg_it = [msg]

    return msg_it


def construct_email(cfg, query, location, posts):
    """Construct an email message.

    Args:
        cfg: email configuration.
        query: search term used.
        location: location to search for jobs.
        posts: dictionary containing new job listings.

    Returns:
        message: string containing the email message.
    """
    nposts = len(posts)

    # unpack required variables
    user, send_to = cfg['email_from'], cfg['email_to']

    # unpack optional variables
    try:
        name = cfg['name']
    except KeyError:
        # if the `name` key isn't present, then use the first part of the
        # email address to address recipient.
        name = cfg['email_to'].split('@')[0]

    try:
        sender_name = cfg['sender_name']
    except KeyError:
        sender_name = cfg['email_from']

    try:
        signature = cfg['signature']
    except KeyError:
        signature = ''

    # some temporary variables to correct grammar in the email message
    was_were = 'was' if nposts == 1 else 'were'
    is_are = 'is' if nposts == 1 else 'are'
    job_s = 'job' if nposts == 1 else 'jobs'
    listing_s = 'listing' if nposts == 1 else 'listings'

    subject = f'Job opportunities: {nposts} new {job_s} posted'
    description = '{}. {jobtitle} @ {company}\nLink: {url}\nLocation: {location}\nSnippet: {desc}\n'
    posts_content = '\n'.join(description.format(i+1, **p) for i, p in enumerate(posts.values()))

    s = (
        f'From: {sender_name} <{user}>\n'
        f'To: {send_to}\n'
        f'Subject: {subject}\n'
        f'Hello {name},\n\n'
        f'There {is_are} {nposts} new job {listing_s} to review.\n'
        f'The following job {listing_s} {was_were} found for {repr(query)} in {repr(location)}:\n\n'
        f'{posts_content}\n'
        f'{signature}'
    )

    return s


def indeed_api_request(params):
    """Performs an API request and returns results.

    Args:
        params: dictionary with search parameters.

    Returns:
        posts: a generator containing dictionaries.

    Raises:
        json.decoder.JSONDecodeError: may be raised if we
            get a malformed response from the API.
    """
    complete_result = False

    while not complete_result:
        url = build_url(INDEED_BASE_URL, params)

        with urllib.request.urlopen(url) as u:
            r = u.read().decode('utf-8')

        response = json.loads(r)

        if 'error' in response:
            raise IndeedAuthenticationError('Invalid Indeed publisher key provided.')

        for result in response['results']:
            yield {
                result['jobkey']:
                    {
                        'jobtitle': result['jobtitle'],
                        'company': result['company'],
                        'date_created': result['date'],
                        'location': result['formattedLocation'],
                        'url': result['url'].split('&')[0],
                        'lat': result['latitude'],
                        'lon': result['longitude'],
                        'desc': result['snippet'],
                    }
                }

        if response['end'] >= response['totalResults']:
            complete_result = True

        # update results start in order to get to the next page
        params['start'] += INDEED_API_LIMIT


def email_notify(cfg, posts, query, location):
    """Notify recipient of new postings.

    Args:
        cfg: email section from configuration file.
        posts: new posts since the last notification
        query: query from `indeed` section of config file
        location: location from `indeed` section of config file
    """
    user = cfg['email_from']
    password = cfg['password']

    message = construct_email(cfg, query, location, posts)
    msg = email.message_from_string(message)
    msg.set_charset('utf-8')

    try:
        send_email(user, password, msg)
    except smtplib.SMTPAuthenticationError as e:
        raise EmailAuthenticationError(
            'Email authentication error. Please check entries for `email_from` '
            'and `password` in your configuration file.'
        )


def send_email(user, password, msg):
    """Send an email.

    Args:
        user: account of sender
        password: password of sender
        msg: email.message.Message object

    Raises:
        smtplib.SMTPAuthenticationError:
            535, b'5.7.8 Username and Password not accepted.
    """
    with smtplib.SMTP('smtp.gmail.com', 587) as smtp:
        smtp.ehlo()  # success 250
        smtp.starttls()  # success 220
        smtp.login(user, password)  # success 235 smtplib.SMTPAuthenticationError
        smtp.send_message(msg)  # empty dict is a success


def slack_notify(cfg, posts):
    """Post a message to a Slack channel.

    Args:
        cfg: configuration for Slack account.
        msg_it: message(s) to be sent. May be a list
            or a generator.

    Raises:
        SlackCfgError: raised if we get a bad response.
    """
    msg_it = construct_slack_message(posts)

    token = cfg['token']
    channel = cfg['channel']

    sc = SlackClient(token)

    # https://api.slack.com/methods/chat.postMessage
    # slack_errors = {
    #     'not_authed': 'No authentication token provided.',
    #     'invalid_auth': 'Invalid authentication token.',
    #     'account_inactive': 'Authentication token is for a deleted user or team.',
    #     'no_text': 'No message text provided',
    #     'not_in_channel': 'Cannot post user messages to a channel they are not in.',
    #     'channel_not_found': 'Value passed for channel was invalid.',
    # }

    r = sc.api_call('api.test')
    if not r['ok']:
        reason = r['error']
        raise SlackCfgError(f'ERROR: {reason}')

    for m in msg_it:
        sc.api_call(
            'chat.postMessage',
            text=m,
            channel=channel,
            icon_emoji=':robot_face:',
        )


def jobnotify(cfg_filename=PATH_TO_CFG, database_dir=DB_DIR):
    """Main entry point for the script"""
    # TODO: if config does not exist perhaps populate with defaults
    if not os.path.isfile(cfg_filename):
        raise FileNotFoundError(f'Configuration file {repr(cfg_filename)} does not exist.')

    # load all configuration files
    cfgs = get_section_configs(cfg_filename)

    # the `indeed` section is the first section in the list
    indeed_cfg = cfgs[0]

    params = {
        'publisher': indeed_cfg['key'],  # publisher ID
        'q': indeed_cfg['query'],  # query
        'l': indeed_cfg['location'],  # location (city, state, region)
        'radius': indeed_cfg['radius'],  # distance from search location 'as the crow flies'
        'jt': 'fulltime',  # job-type: 'fulltime', 'parttime, 'contract', 'temporary', 'internship'
        'limit': 25,  # max number of results per query - max 25
        'fromage': 10,  # number of days back to search
        'start': 0,  # start results at this search number
        'highlight': 0,  # bold search term in snippet
        'latlong': 1,  # return latitude and longitude
        'co': indeed_cfg['country'],  # search within this country
        'v': 2,  # version of the API - should be 2
        'format': 'json',  # response format
    }

    logging.debug(params)

    query, loc = get_sanitised_params(params['q'], params['l'])

    db_name = f'{query}_{loc}.json'
    db_path = os.path.join(database_dir, db_name)
    db = load_json_db(db_path)
    logging.info('Load JSON database %r', db_path)

    # get all listings from the indeed API
    all_posts = indeed_api_request(params)

    # build a list of posts that we haven't seen before
    posts = {k: v for d in all_posts for k, v in d.items() if k not in db}
    logging.info('len(posts)=%d', len(posts))

    if posts:
        # send the notification
        notify(cfgs, posts)

        # update our existing database
        db.update(posts)

        logging.info('Write JSON database %r', db_path)

        write_json_db(db, db_path)
    else:
        logging.info('No new positions since last notification.')


def notify(cfgs, posts):
    """Generic notification function."""
    # assuming cfgs is a list of configs
    # if someone modifies the layout of the cfg file we're in trouble!
    indeed, email, slack, notify_via = cfgs

    if notify_via.getboolean('slack'):
        slack_notify(slack, posts)
        logging.info('Slack message sent with %d listings(s)', len(posts))

    if notify_via.getboolean('email'):
        query = indeed.get('query')
        location = indeed.get('location')
        email_notify(email, posts, query, location)
        logging.info('Email sent with %d listings(s).', len(posts))



def main():
    """Main entry point for this utility."""
    app_data_dir = os.path.join(os.path.expanduser('~'), '.jobnotify')

    # process command line arguments
    # if no command line args are passed sys.argv[1:] == []
    args = process_args(sys.argv[1:])

    if args.config:
        initial_setup(app_data_dir)
        print(f'App data directory created: {app_data_dir}')
        sys.exit()

    if args.verbose:
        # setup basic logging to file
        logging.basicConfig(
            filename=f'.{os.path.splitext(os.path.basename(__file__))[0]}.log',
            format='%(asctime)s %(message)s',
            level=logging.DEBUG
        )
    else:
        # disable logging
        logging.disable(logging.CRITICAL)

    logging.debug('args: %r', args)

    # if the app data directory has not been set up before then do this now
    if not os.path.isdir(app_data_dir):
        initial_setup(app_data_dir)
        logging.info('Created app data directory: %r', app_data_dir)

    try:
        jobnotify(args.file, DB_DIR)
    except (
            ConfigurationFileError,
            DuplicateOptionError,
            FileNotFoundError,
            json.decoder.JSONDecodeError,
            ) as e:
        print(f'ERROR: {e}')
        logging.exception(e)


if __name__ == '__main__':
    sys.exit(main())



import argparse
import base64
import configparser
import json
import logging
import os
from pkg_resources import resource_filename
import shutil

from .exceptions import (
    BlankKeyError,
    NotificationsNotConfiguredError,
    RequiredKeyMissingError,
    SectionNotFoundError,
)


class EmailMatch:
    """Helper class to compare properties of emails."""
    def __init__(self, expected):
        self.expected = expected

    def __eq__(self, other):
        return (
            self.expected['from'] == other['from'] and
            self.expected['to'] == other['to'] and
            self.expected['subject'] == other['subject'] and
            self.expected['message'] == base64.b64decode(other.get_payload()).decode('utf-8')
        )

    def __ne__(self, other):
        """Define a non-equality test"""
        return not self.__eq__(other)

    def __repr__(self):
        return (
            f'From: {self.expected["from"]}\n'
            f'To: {self.expected["to"]}\n'
            f'Subject: {self.expected["subject"]}\n'
            f'Message: {self.expected["message"]}'
        )


def get_sanitised_params(q, l):
    """Remove spaces from params and make lowercase."""
    # TODO: Tidy this string replacement
    query = q.lower().replace(' ', '_').replace('-', '_')
    loc = l.lower().replace(' ', '_').replace('-', '_')

    return query, loc


def load_json_db(path_to_db):
    """Load and return a JSON database.

    Return a JSON database if it exists, otherwise return
    an empty dict.

    Args:
        path_to_db: path to the database.

    Returns:
        db: JSON database, or empty dict

    """
    if not os.path.isfile(path_to_db):
        db = {}
    else:
        with open(path_to_db, 'r') as f:
            db = json.load(f)

    return db


def load_cfg(fname, section, required):
    """Load `section` from configuration file.

    Args:
        section: the section from the configuration file to return
        required: a list of required keys for this section

    Returns:
        dict-like section from configuration file.

    Raises:
        ConfigurationFileError: if `section` missing from file, or
            if required keys are missing.
        configparser.DuplicateOptionError: if there is a duplicate
            key in any section.
    """
    cfg = configparser.ConfigParser()
    cfg.read(fname)

    try:
        cfg_section = cfg[section]
    except KeyError:
        raise SectionNotFoundError(
            f'Section {repr(section)} not found in configuration file.'
        )

    # if the section is missing any of the required keys
    # set-based operation
    if not cfg_section.keys() >= required:
        missing_keys = required - cfg_section.keys()
        raise RequiredKeyMissingError(
            f'Key(s) {missing_keys} keys missing from configuration file.'
        )

    # check that there are no blank fields in the config file
    blank_keys = []
    for k, v in cfg_section.items():
        if not v:
            blank_keys.append(k)

    # TODO: Handle grammar a little better by checking len(blank_keys)
    if blank_keys:
        raise BlankKeyError(f'Key(s) {blank_keys} in {repr(section)} section are blank.')

    return cfg_section


def get_section_configs(filename):
    """Returns sections from configuration file.

    Args:
        filename: configuration filename

    Raises:
        NotificationsNotConfiguredError: if all options in
            `notify_via` section of configuration file are
            set to `false`.

    Returns:
        list of configuration sections.
    """
    # TODO: Can this be turned into a generator?
    cfgs = []

    # TODO: pass in this structure
    section_reqs = [
        ('indeed', {'key', 'query', 'location', 'country', 'radius'}),
        ('email', {'email_from', 'email_to', 'password'}),
        ('slack', {'token', 'channel'}),
        ('notify_via', {'email', 'slack'}),
    ]

    # TODO: Perhaps handle errors at higher level?
    for section, requirements in section_reqs:
        cfgs.append(load_cfg(filename, section, requirements))

    # `notify_via` is the last cfg section added above
    nv = cfgs[-1]

    if not any(nv.getboolean(k) for k in nv.keys()):
        raise NotificationsNotConfiguredError(
            'Notifications must be set for at least one option.'
        )

    indeed = cfgs[0]

    # check that the value passed for radius is valid
    try:
        indeed.getfloat('radius')
    except ValueError:
        print(
            f'WARNING: Bad value for `radius` in `[indeed]` section ({indeed["radius"]}). '
            'Defaulting to radius = 10.'
        )
        logging.info(
            'Bad value (%r) passed for `radius`. Setting radius = 10.', indeed['radius']
        )
        indeed['radius'] = '10'

    return cfgs


def write_json_db(db, path_to_db):
    """Write `db` to file."""
    with open(path_to_db, 'w') as f:
        json.dump(db, f, indent=2, sort_keys=True)


def initial_setup(app_data_dir):
    """Create files needed for `jobnotify` application."""
    sample_config_fn = resource_filename(__name__, 'jobnotify.config.sample')
    config_fn = os.path.join(app_data_dir, 'jobnotify.config')
    database_dir = os.path.join(app_data_dir, 'databases')

    # application directory does not exist - create it
    if not os.path.isdir(app_data_dir):
        os.mkdir(app_data_dir)
        os.mkdir(database_dir)
        shutil.copy(sample_config_fn, config_fn)
    else:
        if not os.path.isdir(database_dir):
            os.mkdir(database_dir)

        # check that the configuration file exists
        if not os.path.exists(config_fn):
            shutil.copy(sample_config_fn, config_fn)


def process_args(
    args=None,
    *,
    path_to_cfg=os.path.join(os.path.expanduser('~'), '.jobnotify', 'jobnotify.config')
):
    """Parse and process command line arguments.

    Args:
        args: command line arguments
        path_to_cfg: path to the default configuration file.

    Returns:
        args: argparse.Namespace object with attributes of
            command line arguments.
    """
    if args is None:
        args = []

    parser = argparse.ArgumentParser(
        description='Send notifications about new job listings.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '-v',
        '--verbose',
        help='enable logging to file',
        action='store_true',
    )
    parser.add_argument(
        '-c',
        '--config',
        help='set up application data directory',
        action='store_true',
    )
    parser.add_argument(
        '-f',
        '--file',
        help='path to configuration file',
        default=path_to_cfg,
    )

    return parser.parse_args(args)



from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()


setup(
    name='jobnotify',
    version='0.1.0',
    description='Send notifications about new job listings.',
    long_description=readme,
    author='Matthew McKenna',
    author_email='mattheweb.mckenna@gmail.com',
    url='https://github.com/matthewmckenna/jobnotify',
    license=license,
    packages=find_packages(exclude=('tests', 'docs', 'htmlcov')),
    entry_points={
        'console_scripts': [
            'jobnotify = jobnotify.jobnotify:main',
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='jobnotify notify email slack job',
    include_package_data=True,
    install_requires=['slackclient>=1.0.5']
)
